from enum import Enum


class FilterOperator(str, Enum):
    """
    class representing Filter operator
    """

    eq = "eq"
    neq = "neq"
    gt = "gt"
    lt = "lt"
    gte = "gte"
    lte = "lte"
    like = "like"
    in_ = "in"



"""Module to represent schema """

from __future__ import annotations
import dataclasses


@dataclasses.dataclass
class Relationship:
    """
    Class to hold a relationship between two tables
    """

    name: str
    table1: str
    field1: str
    table2: str
    field2: str


@dataclasses.dataclass
class UserFriendlyMappings:
    """
    User friendly Names
    """

    fields: list[str]
    metrics: list[str]
    aggregate_metrics: list[str]



from db_chat.sql_builder.FilterOperator import FilterOperator


import dataclasses


@dataclasses.dataclass
class Filter:
    """
    class to represent a filter
    """

    field: str
    operator: FilterOperator
    value: str



"""This is a Sample Python file."""
from __future__ import annotations
from db_chat.sql_builder.Filter import Filter
from db_chat.sql_builder.FilterOperator import FilterOperator
from db_chat.sql_builder.Query import Query
from db_chat.sql_builder.SortOrder import SortOrder

from db_chat.sql_builder.mappings import Relationship


def get_operator_sql(op_type: FilterOperator):
    """
    Get the operator
    """
    operators = {"eq": "=", "neq": "<>", "gt": ">", "lt": "<", "gte": ">=", "lte": "<=", "like": "like", "in": "in"}

    return operators.get(f"{op_type.value}", "")


# TODO: this is rubbish, fix it
def _get_filter_rhs(filter_obj: Filter):
    # TODO: Prevent sql injection
    if filter_obj.operator == FilterOperator.in_:
        val = ""
        # check if the value is a list of strings or a list of ints

        if isinstance(filter_obj.value, list):
            if len(filter_obj.value) > 0:
                if isinstance(filter_obj.value[0], str):
                    for v in filter_obj.value:
                        val += f"'{v}',"
                    val = val[:-1]
                    return f"({val})"

                elif isinstance(filter_obj.value[0], int):
                    for v in filter_obj.value:
                        val += f"{v},"
                    val = val[:-1]
                    return f"({val})"
                else:
                    for v in filter_obj.value:
                        val += f"'{str(v)}',"
                    val = val[:-1]
                    return f"({val})"
            else:
                return f"({val})"

        return f"({val})"
    if filter_obj.operator == FilterOperator.like:
        return f"'%{filter_obj.value}%'"
    else:
        return f"'{filter_obj.value}'"


class SQLBuilder:
    """
    The SQL Builder which builds and emits the SQL
    """

    def __init__(self, table_mapping, field_mappings, relationships):
        self.table_mapping = table_mapping
        self.field_mappings = field_mappings
        self.relationships = relationships

    def build_sql(self, query: Query) -> str:
        """
        Build out the sql
        """
        table = query.table
        fields = query.fields
        filters = query.filters
        sort = query.sort
        limit = query.limit
        offset = query.offset

        # Start with the SELECT part
        sql = self._build_select_clause(table, fields)

        # Resolve relationships (JOINs)
        sql = self._build_joins_and_from_clause(table, fields, filters, sql)

        # WHERE clause
        sql = self._build_where_condition(table, filters, sql)

        # GROUP BY clause for aggregate fields
        sql = self._build_group_by(table, fields, sql)

        # HAVING clause for aggregate filters
        sql = self._build_having(table, filters, sql)

        # Sort order
        if sort:
            sql += self._build_sort_order(table, sort)

        # Offset
        if offset:
            sql += f" OFFSET {offset}"

        # Limit
        if limit:
            sql += f" LIMIT {limit}"

        return sql

    def _build_select_clause(self, root_table, fields):
        select_fields = []
        for field in fields:
            select_field = self._resolve_field_name_for_select(root_table, field)
            select_fields.append(select_field)

        sql = f"SELECT {', '.join(select_fields)} FROM {self.table_mapping[root_table]}"
        return sql

    def _resolve_field_name_for_select(self, root_table: str, field: str) -> str:
        field_detail = self.field_mappings[root_table].get(field, None)
        if not field_detail:
            raise ValueError(f"Unknown field: {field}")

        if isinstance(field_detail, dict) and "relationships" in field_detail and "aggregate" not in field_detail:
            # get last table in the relationship
            alias = field_detail["relationships"][-1]
            return f"{alias}.{field_detail['field']}"
        # TODO: handle aggregate fields in join tables, usually it doesn't make sense to have agg fields in join tables
        elif isinstance(field_detail, dict) and "aggregate" in field_detail:
            return self._add_aggregate_field(root_table, field, field_detail)

        else:
            return f"{self.table_mapping[root_table]}.{field_detail} AS {field}"

    def _add_aggregate_field(self, root_table, field, field_detail):
        agg_function = field_detail["aggregate"]

        if agg_function == "listagg":
            return f"listagg(DISTINCT {field_detail['field']}, ',') AS {field}"

        return f"{agg_function}({self.table_mapping[root_table]}.{field_detail['field']}) AS {field}"

    def _build_joins_and_from_clause(self, root_table, fields, filters: list[Filter], sql):
        mappings = self.field_mappings[root_table]
        all_relationships = []

        all_fields = fields + [filter.field for filter in filters]

        for field in all_fields:
            mapping = mappings.get(field, {})
            if isinstance(mapping, dict) and "relationships" in mapping:
                all_relationships.append(mapping["relationships"])

        added_joins = set()
        for relationship_objs in all_relationships:
            prev_relationship = None
            relationship_name: str
            for relationship_name in relationship_objs:
                if relationship_name not in added_joins:
                    rel: Relationship
                    for rel in self.relationships:
                        if rel.name == relationship_name:
                            join_from = prev_relationship if prev_relationship else self.table_mapping[rel.table1]
                            sql += f" JOIN {self.table_mapping[rel.table2]}  AS {rel.name}"
                            sql += f" ON {join_from}.{rel.field1} = {rel.name}.{rel.field2}"
                            added_joins.add(rel.name)
                prev_relationship = relationship_name

        return sql

    def _resolve_field_name(self, root_table: str, field: str) -> str:
        field_detail = self.field_mappings[root_table].get(field, None)
        if not field_detail:
            raise ValueError(f"Unknown field: {field}")

        if isinstance(field_detail, dict) and "relationships" in field_detail:
            # get last table in the relationship
            alias = field_detail["relationships"][-1]
            return f"{alias}.{field_detail['field']}"

        # TODO: handle aggregate fields in join tables, usually it doesn't make sense to have agg fields in join tables
        elif isinstance(field_detail, dict) and "aggregate" in field_detail:
            return f"{field_detail['field']}"

        else:
            return f"{self.table_mapping[root_table]}.{field_detail}"

    def _build_where_condition(self, root_table, filters, sql):
        where_conditions = []
        filter_obj: Filter
        for filter_obj in filters:
            if "aggregate" not in self.field_mappings[root_table][filter_obj.field]:
                field_name = self._resolve_field_name(root_table, filter_obj.field)
                operator = get_operator_sql(filter_obj.operator)
                field_value = _get_filter_rhs(filter_obj)
                where_conditions.append(f"{field_name} {operator} {field_value}")

        if where_conditions:
            sql += f" WHERE {' AND '.join(where_conditions)}"

        return sql

    def _build_having(self, root_table, filters: list[Filter], sql):
        having_conditions = []
        for filter_obj in filters:
            if "aggregate" in self.field_mappings[root_table][filter_obj.field]:
                having_conditions.append(
                    f"{filter_obj.field} {get_operator_sql(filter_obj.operator)} {_get_filter_rhs(filter_obj)}"
                )

        if having_conditions:
            sql += f" HAVING {' AND '.join(having_conditions)}"

        return sql

    def _build_group_by(self, root_table, fields, sql):
        group_fields = [f for f in fields if "aggregate" not in self.field_mappings[root_table][f]]
        if any("aggregate" in self.field_mappings[root_table][f] for f in fields):
            group_by_fields = [self._resolve_field_name(root_table, field) for field in group_fields]
            sql += f" GROUP BY {', '.join(group_by_fields)}" if any(group_by_fields) else ""
        return sql

    def _build_sort_order(self, root_table, sort: SortOrder):
        if sort is None or sort.field is None:
            return ""
        if self.field_mappings[root_table][sort.field]:
            sort_dir = sort.direction.value if hasattr(sort.direction, "value") else sort.direction
            return f" ORDER BY {sort.field} {sort_dir}"
        else:
            raise ValueError(f"Unknown field: {sort.field}")



from db_chat.sql_builder.sql_builder import (
    SQLBuilder,
    Filter,
    Relationship,
    SortOrder,
    Query as SQLQuery,
    FilterOperator,
)

# Sample Configuration for the Test Cases
table_mapping = {
    "distribution_metrics": "viewership_metrics_day_level",
    "channel": "DW_DIM.channel",
    "content_playout_metrics": "CONTENT_PLAYOUT_METRICS_VIEW",
    "content_playout": "DW_DIM.CONTENT_PLAYOUT",
    "delivery": "DW_DIM.DELIVERY",
    "CUSTOMER": "DW_DIM.CUSTOMER",
    "content": "DW_DIM.AMAGI_ASSET",
    "genre": "DW_DIM.GENRE",
    "genre_mappings": "DW_DIM.GENRE_MAPPINGS",
}

field_mappings = {
    "distribution_metrics": {
        "channel_id": "dim_channel_id",
        "channel_name": {"relationships": ["distribution_channel"], "field": "name"},
        "unique_viewers": "metric_unique_viewers",
        "viewership_minutes": "METRIC_VIEWERSHIP_MINUTES",
        "avg_viewership_minutes": {"aggregate": "SUM", "field": "METRIC_VIEWERSHIP_MINUTES"},
    },
    "channel": {"id": "id", "name": "name"},
    "content_playout_metrics": {
        "unique_viewers": "METRIC_UNIQUE_VIEWERS",
        "content_playout_id": "DIM_CONTENT_PLAYOUT_ID",
        "content_title": {
            "relationships": ["content_playout_metrics_content_playout", "content_playout_content"],
            "field": "title",
        },
        "content_name": {
            "relationships": ["content_playout_metrics_content_playout", "content_playout_content"],
            "field": "name",
        },
        "genre": {
            "aggregate": "listagg",
            "field": "genre.name",
            "relationships": [
                "content_playout_metrics_content_playout",
                "content_playout_content",
                "genre_mappings",
                "genre",
            ],
        },
    },
    "content_playout": {"ID": "ID", "DELIVERY_ID": "DELIVERY_ID", "CONTENT_SOURCE_ASSET_ID": "CONTENT_SOURCE_ASSET_ID"},
    "delivery": {"ID": "ID", "CHANNEL_ID": "CHANNEL_ID"},
    "CUSTOMER": {"AMGID": "AMGID", "NAME": "NAME", "CUSTOMER_TYPE": "CUSTOMER_TYPE"},
    "content": {"NAME": "NAME", "TITLE": "TITLE"},
    "genre": {"NAME": "NAME"},
    "genre_mappings": {},
}

relationships = [
    Relationship(
        name="distribution_channel",
        table1="distribution_metrics",
        field1="dim_channel_id",
        table2="channel",
        field2="id",
    ),
    Relationship(
        name="content_playout_metrics_content_playout",
        table1="content_playout_metrics",
        field1="DIM_CONTENT_PLAYOUT_ID",
        table2="content_playout",
        field2="ID",
    ),
    Relationship(
        name="content_playout_content",
        table1="content_playout",
        field1="CONTENT_SOURCE_ASSET_ID",
        table2="content",
        field2="ID",
    ),
    Relationship(
        name="genre_mappings", table1="content_playout_content", field1="id", table2="genre_mappings", field2="asset_id"
    ),
    Relationship(name="genre", table1="genre_mappings", field1="GENRE_ID", table2="genre", field2="ID"),
]

builder = SQLBuilder(table_mapping, field_mappings, relationships)


def test_simple_field_selection():
    query = {"table": "distribution_metrics", "fields": ["channel_id"]}

    query = SQLQuery(**query)

    # query.fields = query_json["fields"]
    # query.table = query_json["table"]
    # query.fields =
    sql = builder.build_sql(query)
    assert sql == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id FROM viewership_metrics_day_level"


def test_get_channel_name():
    query = {"table": "distribution_metrics", "fields": ["channel_id", "channel_name"]}
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id, distribution_channel.name FROM viewership_metrics_day_level JOIN DW_DIM.channel  AS distribution_channel ON viewership_metrics_day_level.dim_channel_id = distribution_channel.id"
    )


def test_aggregate_field_selection():
    query = {"table": "distribution_metrics", "fields": ["channel_id", "avg_viewership_minutes"]}
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id, SUM(viewership_metrics_day_level.METRIC_VIEWERSHIP_MINUTES) AS avg_viewership_minutes FROM viewership_metrics_day_level GROUP BY viewership_metrics_day_level.dim_channel_id"
    )


def test_nested_join():
    query = {"table": "content_playout_metrics", "fields": ["unique_viewers", "content_title", "content_name"]}
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT CONTENT_PLAYOUT_METRICS_VIEW.METRIC_UNIQUE_VIEWERS AS unique_viewers, content_playout_content.title, content_playout_content.name FROM CONTENT_PLAYOUT_METRICS_VIEW JOIN DW_DIM.CONTENT_PLAYOUT  AS content_playout_metrics_content_playout ON CONTENT_PLAYOUT_METRICS_VIEW.DIM_CONTENT_PLAYOUT_ID = content_playout_metrics_content_playout.ID JOIN DW_DIM.AMAGI_ASSET  AS content_playout_content ON content_playout_metrics_content_playout.CONTENT_SOURCE_ASSET_ID = content_playout_content.ID"
    )


def test_filter_on_normal_field():
    query = {
        "table": "content_playout_metrics",
        "fields": ["unique_viewers"],
        "filters": [Filter(field="unique_viewers", operator=FilterOperator.gt, value="100")],
    }
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT CONTENT_PLAYOUT_METRICS_VIEW.METRIC_UNIQUE_VIEWERS AS unique_viewers FROM CONTENT_PLAYOUT_METRICS_VIEW WHERE CONTENT_PLAYOUT_METRICS_VIEW.METRIC_UNIQUE_VIEWERS > '100'"
    )


def test_filter_on_join_field():
    query = {
        "table": "content_playout_metrics",
        "fields": ["unique_viewers", "content_title"],
        "filters": [Filter(field="content_title", operator=FilterOperator.eq, value="Liz Pede Bis")],
    }
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT CONTENT_PLAYOUT_METRICS_VIEW.METRIC_UNIQUE_VIEWERS AS unique_viewers, content_playout_content.title FROM CONTENT_PLAYOUT_METRICS_VIEW JOIN DW_DIM.CONTENT_PLAYOUT  AS content_playout_metrics_content_playout ON CONTENT_PLAYOUT_METRICS_VIEW.DIM_CONTENT_PLAYOUT_ID = content_playout_metrics_content_playout.ID JOIN DW_DIM.AMAGI_ASSET  AS content_playout_content ON content_playout_metrics_content_playout.CONTENT_SOURCE_ASSET_ID = content_playout_content.ID WHERE content_playout_content.title = 'Liz Pede Bis'"
    )


def test_filter_on_aggregate_having_clause():
    query = {
        "table": "distribution_metrics",
        "fields": ["channel_id", "avg_viewership_minutes"],
        "filters": [Filter(field="avg_viewership_minutes", operator=FilterOperator.gt, value="100")],
    }
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id, SUM(viewership_metrics_day_level.METRIC_VIEWERSHIP_MINUTES) AS avg_viewership_minutes FROM viewership_metrics_day_level GROUP BY viewership_metrics_day_level.dim_channel_id HAVING avg_viewership_minutes > '100'"
    )


def test_sort_on_aggregates():
    query = {
        "table": "distribution_metrics",
        "fields": ["channel_id", "avg_viewership_minutes"],
        "filters": [Filter(field="avg_viewership_minutes", operator=FilterOperator.gt, value="100")],
        "sort": SortOrder(field="avg_viewership_minutes", direction="DESC"),
    }
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id, SUM(viewership_metrics_day_level.METRIC_VIEWERSHIP_MINUTES) AS avg_viewership_minutes FROM viewership_metrics_day_level GROUP BY viewership_metrics_day_level.dim_channel_id HAVING avg_viewership_minutes > '100' ORDER BY avg_viewership_minutes DESC"
    )


def test_offset():
    query = {"table": "distribution_metrics", "fields": ["channel_id"], "offset": 10}
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id FROM viewership_metrics_day_level OFFSET 10"
    )


def test_limit():
    query = {"table": "distribution_metrics", "fields": ["channel_id"], "limit": 10}
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id FROM viewership_metrics_day_level LIMIT 10"
    )


def test_in_filter():
    query = {
        "table": "distribution_metrics",
        "fields": ["channel_id"],
        "filters": [Filter(field="channel_id", operator=FilterOperator.in_, value=["1", "2", "3"])],
    }
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT viewership_metrics_day_level.dim_channel_id AS channel_id FROM viewership_metrics_day_level WHERE viewership_metrics_day_level.dim_channel_id in ('1','2','3')"
    )


def test_list_agg():
    query = {"table": "content_playout_metrics", "fields": ["content_title", "genre"], "filters": [], "limit": 10}
    query = SQLQuery(**query)
    sql = builder.build_sql(query)
    assert (
        sql
        == "SELECT content_playout_content.title, listagg(DISTINCT genre.name, ',') AS genre FROM CONTENT_PLAYOUT_METRICS_VIEW JOIN DW_DIM.CONTENT_PLAYOUT  AS content_playout_metrics_content_playout ON CONTENT_PLAYOUT_METRICS_VIEW.DIM_CONTENT_PLAYOUT_ID = content_playout_metrics_content_playout.ID JOIN DW_DIM.AMAGI_ASSET  AS content_playout_content ON content_playout_metrics_content_playout.CONTENT_SOURCE_ASSET_ID = content_playout_content.ID JOIN DW_DIM.GENRE_MAPPINGS  AS genre_mappings ON content_playout_content.id = genre_mappings.asset_id JOIN DW_DIM.GENRE  AS genre ON genre_mappings.GENRE_ID = genre.ID GROUP BY content_playout_content.title LIMIT 10"
    )


# def test_filter_on_normal_field():
#     sql = builder.build_sql('Order', ['ID'], [Filter(field='ID', operator='=', value='5')])
#     assert sql == "SELECT orders.order_id AS ID FROM orders WHERE orders.order_id = '5'"

# def test_sort_order():
#     sql = builder.build_sql('Order', ['ID'], sort=SortOrder(field='ID', direction='DESC'))
#     assert sql == "SELECT orders.order_id AS ID FROM orders ORDER BY orders.order_id DESC"

# def test_limit():
#     sql = builder.build_sql('Order', ['ID'], limit=10)
#     assert sql == "SELECT orders.order_id AS ID FROM orders LIMIT 10"

# You can continue to add more test cases as necessary based on the functionality and edge cases you want to cover.


# def test_simple_field_selection():
#     sql = builder.build_sql('Order', ['ID'])
#     assert sql == "SELECT orders.order_id AS ID FROM orders"

# def test_related_field_selection():
#     sql = builder.build_sql('Order', ['ID', 'UserName'])
#     assert sql == "SELECT orders.order_id AS ID, users.user_name AS UserName FROM orders JOIN users ON orders.user_id = users.id"

# def test_aggregate_field_selection():
#     sql = builder.build_sql('Order', ['TotalAmount'])
#     assert sql == "SELECT SUM(orders.amount) AS TotalAmount FROM orders"

# def test_filter_on_aggregate():
#     sql = builder.build_sql('Order', ['TotalAmount'], [Filter(field='TotalAmount', operator='>', value='100')])
#     assert sql == "SELECT SUM(orders.amount) AS TotalAmount FROM orders HAVING SUM(orders.amount) > '100'"

# def test_filter_on_normal_field():
#     sql = builder.build_sql('Order', ['ID'], [Filter(field='ID', operator='=', value='5')])
#     assert sql == "SELECT orders.order_id AS ID FROM orders WHERE orders.order_id = '5'"

# def test_sort_order():
#     sql = builder.build_sql('Order', ['ID'], sort=SortOrder(field='ID', direction='DESC'))
#     assert sql == "SELECT orders.order_id AS ID FROM orders ORDER BY orders.order_id DESC"

# def test_limit():
#     sql = builder.build_sql('Order', ['ID'], limit=10)
#     assert sql == "SELECT orders.order_id AS ID FROM orders LIMIT 10"

# # You can continue to add more test cases as necessary based on the functionality and edge cases you want to cover.



""" Playground file to just write some code to test some stuff out.  """
from sqlalchemy import ForeignKey, create_engine, Table, Column, Integer, String, MetaData, select, join, alias

# engine = create_engine("sqlite:///example.db", echo=True)
metadata = MetaData()

users = Table(
    "users",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String),
    Column("fullname", String),
)

addresses = Table(
    "addresses",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", None, ForeignKey("users.id")),
    Column("email_address", String, nullable=False),
)

statement = addresses

# Define an alias for the users table
users_alias = alias(users, "address_user")


# Build a join between the users table and the addresses table using the alias
join_condition = users_alias.c.id == addresses.c.user_id
join = join(users_alias, statement, join_condition)

# Select columns from both tables using the alias
stmt = select(users_alias.columns.name.label("user name"), addresses.c.email_address).select_from(join)

print(str(stmt))
# Execute the statement
# with engine.connect() as conn:
#     result = conn.execute(stmt).fetchall()



""" Query Class """
from __future__ import annotations
import dataclasses

from db_chat.sql_builder.Filter import Filter
from db_chat.sql_builder.SortOrder import SortOrder


from enum import Enum


class Functions(str, Enum):
    """
    class representing Filter operator
    """

    SUM = "SUM"
    MIN = "MIN"
    MAX = "MAX"
    AVG = "AVG"
    CURRENT_DATE = "CURRENT_DATE"


@dataclasses.dataclass
class Expression:
    """
    Class to encapsulate a complex field
    """

    func: str
    params: list[str] = dataclasses.field(default_factory=lambda: [])
    label: str = None


@dataclasses.dataclass
class Query:
    """
    Class to encapsulate the SQL query
    """

    table: str
    fields: list[str | Expression] = dataclasses.field(default_factory=lambda: [])
    filters: list[Filter] = dataclasses.field(default_factory=lambda: [])
    group_by: list[str | Expression] = dataclasses.field(default_factory=lambda: [])
    sort: SortOrder = None
    limit: int = None
    offset: int = None



from __future__ import annotations
from db_chat.sql_builder.Filter import Filter
from db_chat.sql_builder.FilterOperator import FilterOperator
from db_chat.sql_builder.Query import Query, Expression, Functions

from db_chat.sql_builder.sqlalchemy_query_builder import Column, SQLAlchemyQueryBuilder, Schema, Table
from db_chat.sql_builder.mappings import Relationship


schema: Schema = Schema(
    tables={
        "users": Table(
            name="users",
            friendly_name="users",
            columns=[
                Column(name="name", friendly_name="name"),
                Column(name="id", friendly_name="id"),
                Column(name="email", friendly_name="email"),
                Column(name="posts", friendly_name="posts", relationships=["post_user"]),
            ],
        ),
        "posts": Table(
            name="posts",
            friendly_name="posts",
            columns=[
                Column(name="id", friendly_name="id"),
                Column(name="user_id", friendly_name="user_id"),
                Column(name="title", friendly_name="title"),
                Column(name="body", friendly_name="body"),
                Column(name="likes", friendly_name="likes"),
                Column(name="user_name", friendly_name="user name", relationships=["post_user"], related_field="name"),
                Column(
                    name="user_email", friendly_name="user email", relationships=["post_user"], related_field="email"
                ),
            ],
        ),
        "comments": Table(
            name="comments",
            friendly_name="comments",
            columns=[
                Column(name="id", friendly_name="id"),
                Column(name="post_id", friendly_name="post_id"),
                Column(name="body", friendly_name="body"),
                Column(
                    name="comment_user_name",
                    friendly_name="user name",
                    relationships=["post_comments", "post_user"],
                    related_field="name",
                ),
            ],
        ),
    },
    relationships=[
        Relationship(name="post_user", table1="posts", field1="user_id", table2="users", field2="id"),
        Relationship(name="post_comments", table1="posts", field1="id", table2="comments", field2="post_id"),
    ],
)

builder = SQLAlchemyQueryBuilder(schema)


def test_simple():
    """
    Simple query
    """

    query = Query(
        table="users",
        fields=["name", "email"],
        filters=[],
        sort=["name"],
        limit=10,
        offset=10,
    )

    Sql = builder.build_query(query=query)
    assert Sql == "SELECT users.name, users.email \nFROM users"


def test_simple_join():
    """
    Simple join
    """

    query = Query(
        table="posts",
        fields=["title", "body", "user_name"],
    )

    Sql = builder.build_query(query=query)
    assert (
        Sql
        == "SELECT posts.title, posts.body, posts_post_user.name AS user_name \nFROM posts JOIN users AS posts_post_user ON posts.user_id = posts_post_user.id"
    )


def test_join_with_multiple_fields():
    """
    simple join with multiple fields
    """

    query = Query(
        table="posts",
        fields=["title", "body", "user_name", "user_email"],
    )
    S = builder.build_query(query=query)
    assert (
        S
        == "SELECT posts.title, posts.body, posts_post_user.name AS user_name, posts_post_user.email AS user_email \nFROM posts JOIN users AS posts_post_user ON posts.user_id = posts_post_user.id"
    )


def test_two_level_join():
    """
    Two level join
    """

    query = Query(
        table="comments",
        fields=["body", "comment_user_name"],
        filters=[],
        sort=["title"],
        limit=10,
        offset=10,
    )

    Sql = builder.build_query(query=query)
    assert (
        Sql
        == "SELECT comments.body, comments_post_comments_post_user.name AS comment_user_name \nFROM comments JOIN posts AS comments_post_comments ON comments.post_id = comments_post_comments.id JOIN users AS comments_post_comments_post_user ON comments_post_comments.user_id = comments_post_comments_post_user.id"
    )


def test_simple_filter():
    """
    Simple filter
    """
    query = Query(
        table="users",
        fields=["name", "email"],
        filters=[Filter(field="name", operator=FilterOperator.eq, value="John")],
        sort=["name"],
        limit=10,
        offset=10,
    )
    sql = builder.build_query(query=query)
    assert sql == "SELECT users.name, users.email \nFROM users \nWHERE users.name = :name_1"


# filters with joins
def test_filter_with_join():
    """
    Filter with join
    """
    query = Query(
        table="posts",
        fields=["title", "body", "user_name"],
        filters=[Filter(field="user_name", operator=FilterOperator.eq, value="John")],
    )
    sql = builder.build_query(query=query)
    assert (
        sql
        == "SELECT posts.title, posts.body, posts_post_user.name AS user_name \nFROM posts JOIN users AS posts_post_user ON posts.user_id = posts_post_user.id \nWHERE posts_post_user.name = :name_1"
    )


# aggregates and functions
def test_simple_aggregate():
    """
    Simple aggregate
    """
    query = Query(
        table="posts",
        fields=[Expression(func="SUM", params=["likes"], label="total_likes")],
        filters=[],
        sort=["title"],
        limit=10,
        offset=10,
    )
    sql = builder.build_query(query=query)
    assert sql == "SELECT sum(posts.likes) AS total_likes \nFROM posts"


# filters with aggregates and functions

# filters with constants and functions


# group by
def test_group_by():
    """
    Group by
    """
    query = Query(
        table="posts",
        fields=["title", "body", "user_name"],
        group_by=["title", "body", "user_name"],
    )
    sql = builder.build_query(query=query)
    assert (
        sql
        == "SELECT posts.title, posts.body, posts_post_user.name AS user_name \nFROM posts JOIN users AS posts_post_user ON posts.user_id = posts_post_user.id GROUP BY posts.title, posts.body, posts_post_user.name"
    )


# Having clause

# or and not and and conditions in joins



"""     Query Builder that builds SQL Alchmey Query """

from __future__ import annotations
from dataclasses import dataclass
import dataclasses
from sqlalchemy import alias, and_, select, table, column, TableClause, func
from db_chat.sql_builder.Filter import Filter
from db_chat.sql_builder.FilterOperator import FilterOperator

from db_chat.sql_builder.Query import Expression, Functions, Query
from db_chat.sql_builder.mappings import Relationship


@dataclass
class Schema:
    """
    Schema
    """

    tables: dict[str, Table]
    relationships: list[Relationship]


@dataclass
class Table:
    """
    Table
    """

    friendly_name: str
    name: str
    columns: list[Column]
    relationships: list[str] = dataclasses.field(default_factory=lambda: [])


@dataclass
class Column:
    """
    Column
    """

    friendly_name: str
    name: str
    relationships: list[str] = dataclasses.field(default_factory=lambda: [])
    related_field: str = None


class SQLAlchemyQueryBuilder:
    """
    Query Builder that builds SQL Alchmey Query
    """

    def __init__(self, schema: Schema):
        self.schema = schema

    def build_query(self, query: Query):
        """
        Build out the sql
        """

        statement = self._build_from_clause(query)

        # select_clause = self._build_select_clause(query)

        # # Start with the SELECT part
        # sql = self._build_select_clause(table, fields)

        # # Resolve relationships (JOINs)
        # sql = self._build_joins_and_from_clause(table, fields, filters, sql)

        # # WHERE clause
        # sql = self._build_where_clause(filters, sql)

        # # ORDER BY clause
        # sql = self._build_order_by_clause(sort, sql)

        # # LIMIT clause
        # sql = self._build_limit_clause(limit, sql)

        # # OFFSET clause
        # sql = self._build_offset_clause(offset, sql)

        sql = str(statement.compile())
        print(sql)
        return sql

    def _build_from_clause(self, query: Query):
        """
        Build the FROM clause
        """

        sa_table = self._get_table_from_mapping(query.table)
        select_columns = []
        joined_paths = []
        joined_tables: dict = {}

        from_clause = self._build_select_clause_for_fields(query, sa_table, select_columns, joined_paths, joined_tables)

        from_clause, filter_clauses = self._build_filter_clause(
            query, sa_table, joined_paths, joined_tables, from_clause
        )

        from_clause, group_by_clauses = self._build_group_by_columns(
            query, sa_table, joined_paths, joined_tables, from_clause
        )

        statement = select(*select_columns).select_from(from_clause)

        if len(filter_clauses) > 0:
            statement = statement.where(and_(*filter_clauses))

        if len(group_by_clauses) > 0:
            statement = statement.group_by(*group_by_clauses)

        return statement

    def _build_group_by_columns(self, query: Query, sa_table: TableClause, joined_paths, joined_tables, from_clause):
        group_by_columns = []
        for group_by_field in query.group_by:
            group_by_column = self._get_field_from_table(query.table, group_by_field)

            if not group_by_column.relationships:
                sa_grouping_column = sa_table.columns.get(group_by_column.name)
            else:
                from_clause, join_to_aliased_table = self._build_join_for_relationship(
                    sa_table, from_clause, joined_paths, joined_tables, group_by_column, query.table, query.table
                )
                sa_grouping_column = join_to_aliased_table.columns.get(group_by_column.related_field)

            group_by_columns.append(sa_grouping_column)
        return from_clause, group_by_columns

    def _build_filter_clause(self, query: Query, sa_table: TableClause, joined_paths, joined_tables, from_clause):
        filter_clauses = []
        for filter_obj in query.filters:
            filter_column = self._get_field_from_table(query.table, filter_obj.field)

            if not filter_column.relationships:
                sa_filter_column = sa_table.columns.get(filter_column.name)
            else:
                from_clause, join_to_aliased_table = self._build_join_for_relationship(
                    sa_table, from_clause, joined_paths, joined_tables, filter_column, query.table, query.table
                )
                sa_filter_column = join_to_aliased_table.columns.get(filter_column.related_field)
            filter_clause = self._get_filter_condition_for_operator(filter_obj, sa_filter_column)
            filter_clauses.append(filter_clause)
        return from_clause, filter_clauses

    def _get_filter_condition_for_operator(self, filter_obj: Filter, sa_filter_column):
        # switch case for operators
        operator = filter_obj.operator.value

        filter_clause = {
            f"{FilterOperator.eq.value}": lambda x, y: x == y,
        }.get(
            operator
        )(sa_filter_column, filter_obj.value)

        return filter_clause

    def _build_select_clause_for_fields(
        self, query: Query, sa_table: TableClause, select_columns: list, joined_paths, joined_tables
    ):
        from_clause = sa_table
        query_field: str | Expression
        for query_field in query.fields:
            field_name: str = query_field
            if isinstance(query_field, Expression):
                sa_expression_function = self._build_clause_for_expression(
                    query, from_clause, query_field, sa_table, joined_paths, joined_tables
                )
                select_columns.append(sa_expression_function)
                continue

            field_column = self._get_field_from_table(query.table, field_name)

            current_table_name = query.table
            current_alias = query.table

            if not field_column.relationships:
                select_columns.append(sa_table.columns.get(field_column.name))

            if field_column.relationships:
                from_clause, join_to_aliased_table = self._build_join_for_relationship(
                    sa_table, from_clause, joined_paths, joined_tables, field_column, current_table_name, current_alias
                )

                select_columns.append(
                    join_to_aliased_table.columns.get(field_column.related_field).label(field_column.name)
                )

        return from_clause

    def _build_clause_for_expression(
        self,
        query: Query,
        from_clause: TableClause,
        expression: Expression,
        sa_table: TableClause,
        joined_paths,
        joined_tables,
    ):
        function_parms = []

        for func_param in expression.params:
            if not self._is_expression(func_param):
                # if not a expression check if it is a field of the table
                is_field_of_table = self._is_field_of_table(func_param, query.table)
                if is_field_of_table:
                    field_column = self._get_field_from_table(query.table, func_param)
                    if not field_column.relationships:
                        function_parms.append(sa_table.columns.get(field_column.name))
                    else:
                        from_clause, join_to_aliased_table = self._build_join_for_relationship(
                            sa_table,
                            from_clause,
                            joined_paths,
                            joined_tables,
                            field_column,
                            query.table,
                            query.table,
                        )

                        function_parms.append(
                            join_to_aliased_table.columns.get(field_column.related_field).label(field_column.name)
                        )
                else:
                    # if not a field of the table then it is a constant
                    function_parms.append(func_param)

        sa_function = self._get_sa_function(expression.func, function_parms, expression.label)

        return sa_function

    def _get_sa_function(self, func_name: str, function_parms: list, label: str):
        sa_function = None
        if func_name == Functions.SUM.value:
            sa_function = func.sum(*function_parms)

        if label:
            sa_function = sa_function.label(label)

        return sa_function

    def _is_expression(self, input_str: str):
        return not input_str.isalnum()

    def _is_field_of_table(self, input_str: str, table_name: str):
        if self._get_field_from_table(table_name, input_str) is not None:
            return True
        return False

    def _build_join_for_relationship(
        self,
        sa_table: TableClause,
        from_clause,
        joined_paths: list,
        joined_tables: dict,
        field_column,
        current_table_name,
        current_alias,
    ):
        previous_table = sa_table
        for relationship_name in field_column.relationships:
            relationship_object: Relationship = self._get_relationship_by_name(relationship_name)
            if relationship_object.table1 == current_table_name:
                target_table_name = relationship_object.table2
                source_field_name = relationship_object.field1
                target_field_name = relationship_object.field2

            elif relationship_object.table2 == current_table_name:
                target_table_name = relationship_object.table1
                source_field_name = relationship_object.field2
                target_field_name = relationship_object.field1

            else:
                raise ValueError(f"Unknown table in relationship: {relationship_name}")

            current_alias += "_" + relationship_object.name

            if current_alias not in joined_paths:
                join_to_table = self._get_table_from_mapping(table_name=target_table_name)
                source_join_column = previous_table.columns.get(source_field_name)
                join_to_aliased_table = alias(join_to_table, current_alias)
                target_join_column = join_to_aliased_table.columns.get(target_field_name)
                join_condition = source_join_column == target_join_column
                from_clause = from_clause.join(join_to_aliased_table, join_condition)
                joined_paths.append(current_alias)
                joined_tables[current_alias] = join_to_aliased_table
            else:
                join_to_aliased_table = joined_tables[current_alias]

            current_table_name = target_table_name
            previous_table = join_to_aliased_table
        return from_clause, join_to_aliased_table

    def _get_relationship_by_name(self, relationship_name):
        return next((rel for rel in self.schema.relationships if rel.name == relationship_name), None)

    def _get_field_from_table(self, table_name, field_name):
        return next(
            (c for c in self.schema.tables[table_name].columns if c.name == field_name),
            None,
        )

    def _get_table_from_mapping(self, table_name: str):
        """
        Get the table from the mapping
        """

        # get the table from the schema
        schema_table = self.schema.tables[table_name]

        # get the columns
        sa_columns = []
        schema_column: Column
        for schema_column in schema_table.columns:
            sa_column = column(schema_column.name)
            sa_columns.append(sa_column)

        sa_table = table(schema_table.name, *sa_columns)

        return sa_table

    def _get_select_fields(self, fields: list[str]):
        """
        Get the select fields
        """

        # get the columns
        sa_columns = []
        field: str
        for field in fields:
            sa_column = column(field)
            sa_columns.append(sa_column)

        return sa_columns
